import logging
import pathlib
import typing

import yaml

# from .abc import AbstractManifest, AbstractStage
from .exc import BuildError

logger = logging.getLogger('buildr')


class ManifestV1:
    """Required properties for a version 1 manifest"""
    def __init__(self, manifest_def):
        self._def = manifest_def
        self._stages = None  # type: Optional[List]
        self._env = None  # type: Optional[List[str]]

    @property
    def stages(self):
        """List of stages to be executed, in order"""
        if self._stages is None:
            logger.debug('Loading stages')
            self._stages = self._def.get('stages', [])
            if self._def.get('prepare'):
                logger.debug('Found prepare, inserting as stage 0')
                if 'prepare' in self._stages:
                    self._stages.remove('prepare')
                self._stages.insert(0, 'prepare')
            logger.debug('Loaded stages: %s', self._stages)
        return self._stages

    @property
    def image(self):
        """Base docker image to run within"""
        return self._def.get('image', 'ubuntu:latest')

    @property
    def env(self):
        """Prepared environmental variables"""
        if self._env is None:
            logger.debug('Loading envvars')
            self._env = []
            env = self._def.get('environment', {})
            if env.get('inherit', False):
                logger.debug('Loading parent envvars')
                for e in os.environ.items():
                    self._env.append('='.join(e))
            for var in env.get('vars', []):
                self._env.append(var)
            logger.debug('Loaded %d envvars', len(self._env))
        return self._env

    def __getitem__(self, value):
        """Access stage definitions as dict keys"""
        return self._def.get(value)


    @staticmethod
    def from_file(project_dir: pathlib.Path):
        """Load the manifest from the project directory,
        erroring on the manifest not being found or the
        project directory not existing.
        :param project_dir: Where the project is on disk
        :returns ManifestV1: The parsed manifest"""
        logger.info('Searching for manifest in %s', project_dir.absolute())
        if not project_dir.exists():
            logger.error('Project directory does not exist.')
            raise BuildError('Project directory does not exist.')

        buildr = None
        for file_name in ['.buildr', '.buildr.yml', '.buildr.yaml']:
            logger.debug('Manifest candidate: %s', file_name)
            candidate = project_dir / file_name
            if not candidate.exists():
                logger.debug('Not found, moving on.')
                continue
            buildr = candidate
            break

        if buildr is None:
            print('Manifest not found, unable to continue')
            raise BuildError('Build manifest not found')
        logger.info('Found manifest: %s', buildr.absolute())

        with open(str(buildr.absolute()), 'r') as f:
            return ManifestV1(yaml.safe_load(f))



"""\
Run the target project's manifest file, emulating a build
run on a CI system.
"""
import contextlib
import logging
import pathlib
import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from typing import Dict, Any, List

from buildr.buildr import Buildr
from buildr.exc import BuildError, BuildFailure
from buildr.manifest import ManifestV1

logger = logging.getLogger('buildr')


def run_manifest(manifest: ManifestV1, target_dir: pathlib.Path, *,
                 docker_sock='unix://var/run/docker.sock',
                 progress_writer=sys.stdout.write,
                 project_meta=None):
    """Run the manifest stages.

    :param manifest: The manifest to run
    :param target_dir: Which directory the source is in that the manifest
                       targets.
    :param docker_sock: Location of the docker engine socket.
    :param progress_writer: Stream to write echoed stdout from the
                            build container to, defaults to stdout."""
    logger.debug('Using docker sock: "%s"', docker_sock)

    with Buildr(target_dir, docker_sock=docker_sock, image=manifest.image,
                env=manifest.env) as buildr:
        for stage_name in manifest.stages:
            logger.info('Running stage "%s"', stage_name)
            stage = manifest[stage_name]
            for script in stage.get('script', []):
                rc = buildr.execute(script, writer=progress_writer)
                if rc != 0:
                    # TODO: get rid of the prepare concept and have a pre/post
                    # each stage instead, optionally.
                    if stage_name == 'prepare':
                        logger.error('Command exited with error, unable to '
                                        'prepare the environment.')
                        raise BuildError('Prepare failed, unable to set up '
                                            'the environment.')
                    else:
                        logger.error('Command exited with error, build '
                                        'failed.')
                        raise BuildFailure('Stage "%s" failed.', stage_name)
            logger.info('Stage "%s" success', stage_name)


def main():
    parser = ArgumentParser(prog='buildr',
                            description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--path', help='Local directory path', type=str,
                        default='.')
    parser.add_argument('--docker-sock', help='Docker sock', type=str,
                        default='unix://var/run/docker.sock')
    parser.add_argument('-d', '--debug', help='Debug logging',
                        action='store_const', dest='loglevel',
                        const=logging.DEBUG, default=logging.WARNING)
    parser.add_argument('-v', '--verbose', help='Be verbose',
                        action='store_const', dest='loglevel',
                        const=logging.INFO)
    args = parser.parse_args()

    logging.basicConfig()
    logger.setLevel(args.loglevel)

    target_dir = pathlib.Path(args.path)
    logger.info('Loading manifest from project directory %s',
                target_dir.absolute())
    manifest = ManifestV1.from_file(target_dir)
    try:
        run_manifest(manifest, target_dir, docker_sock=args.docker_sock)
    except BuildError as e:
        logger.error('Error preparing the environment, most likely unrelated to'
                     ' your code: %s', e)
        raise SystemExit(-1)
    except BuildFailure as e:
        logger.error('Build failed: %s', e)
        raise SystemExit(1)


if __name__ == '__main__':
    with contextlib.suppress(KeyboardInterrupt):
        main()



import logging
import pathlib
import sys

import docker

logger = logging.getLogger('buildr')


class Buildr:
    def __init__(self, project_dir: pathlib.Path, *,
                 docker_sock='unix://var/run/docker.sock', image='docker',
                 env=None):
        """
        Main class that controls the build process, it provides a simple
        API so we can easily abstract out other ways to build if we don't
        end up using docker for them all. It could be cool to have a
        vagrant way.

        :param project_dir: Project directory path
        :param docker_sock: Docker socket
        :param image: Builder image, it must have docker installed in it.
                      Presuming you will be pulling/pushing from private
                      repos, you should have your docker/config.json in
                      the main user's home.
        """
        self.project_dir = project_dir
        self.base_url = docker_sock
        self.image = image
        self.env = env
        if env is None:
            self.env = []

        self.cli = None
        self.container_id = None
        self._cm = False

    def __enter__(self):
        logger.debug('Using docker sock %s', self.base_url)
        self.cli = docker.Client(base_url=self.base_url)
        self._pull_container()
        self.container_id = self._create_container()
        self._start_container()
        self._cm = True

        return self

    def __exit__(self, *args):
        self._close_container()
        self._cm = False

    def execute(self, command: str, writer=sys.stdout.write) -> int:
        """Execute a build command.
        :param command: Command to execute in the shell
        :returns: Exit code (int)"""
        if not self._cm:
            raise ValueError('Buildr must be run as a context manager to'
                             ' ensure all resources are reaped on exit.')

        # TODO: maybe we need to not log the command in case there is
        # something private in there...
        logger.debug('Executing command: "%s"', command)
        # You can emulate this with:
        #   docker exec <container_name> <script>
        exec_ = self.cli.exec_create(self.container_id, command)
        exec_id = exec_['Id']
        for chunk in self.cli.exec_start(exec_id, stream=True):
            try:
                writer(chunk.decode())
            except UnicodeDecodeError:
                pass

        result = self.cli.exec_inspect(exec_id)
        ec = result['ExitCode']
        if ec == 0:
            logger.debug('Command execution success')
        else:
            logger.error('Command execution failed: %r', result)
        return ec

    def _pull_container(self):
        if not ':' in self.image:
            sys.stderr.write('No tag provided on the image, defaulting '
                             'to latest')
            self.image += ':latest'
        for chunk in self.cli.pull(self.image, stream=True):
            try:
                sys.stdout.write(chunk.decode())
            except UnicodeDecodeError:
                pass

    def _create_container(self):
        """Creates the build runner container"""
        # For debug, you can emulate this with:
        #   docker run -it -v /var/run/docker.sock:/var/run/docker.sock -v "${PWD}":/app \
        #   --workdir /app -e <your envvars> <manifest.image> sh
        volumes = ['/app', '/var/run/docker.sock']
        host_config = self.cli.create_host_config(binds=[
            '{}:/app'.format(self.project_dir.absolute()),
            '/var/run/docker.sock:/var/run/docker.sock',
        ])
        logger.debug('Creating container "%s"', self.image)
        container = self.cli.create_container(image=self.image,
                                              command='sh',
                                              detach=True,
                                              environment=self.env,
                                              stdin_open=True,
                                              working_dir='/app',
                                              volumes=volumes,
                                              host_config=host_config)
        return container['Id']

    def _start_container(self):
        """Start the container"""
        logger.debug('Starting container "%s"', self.container_id)
        self.cli.start(self.container_id)

    def _close_container(self):
        """Close, shutdown, and remove the container"""
        logger.debug('Killing and removing container "%s"', self.container_id)
        self.cli.kill(self.container_id)
        self.cli.remove_container(self.container_id)



class BuildError(Exception):
    """Defines an error during the process, this doesn't mean
    anything failed but instead that either a precondition failed
    or a pre-build item."""


class BuildFailure(Exception):
    """Defines a failure in a build step,"""



import abc
import typing


class AbstractStage(abc.ABC):
    """Stage definition"""

    @abc.abstractproperty
    def image(self) -> typing.Optional[str]:
        """Image to use for the stage, if any"""

    @abc.abstractproperty
    def pre(self) -> typing.List[str]:
        """Pre-run work - errors that happen
        during this group count as an error"""

    @abc.abstractproperty
    def run(self) -> typing.List[str]:
        """Main build run, errors that happen
        during this group count as a failure"""

    @abc.abstractproperty
    def post(self) -> typing.List[str]:
        """Post-run work, error that happen
        during this group count as an error"""

    @abc.abstractproperty
    def skip(self) -> bool:
        """Defines if the stage should be skipped
        for some reason."""


class AbstractManifest(abc.ABC):
    """Manifest definition"""

    @abc.abstractproperty
    def stages(self) -> typing.Optional[AbstractStage]:
        """List of stages to run, in order."""

    @abc.abstractproperty
    def image(self) -> str:
        """Image to use for the full run, if any."""

    @abc.abstractproperty
    def env(self) -> typing.Optional[str]:
        """Environment for the image."""



"""\
Simple executor for yaml defined builds, plugs easily into existing CI
systems like Bamboo or Jenkins. Everything is intended to be built fully
isolated in fresh docker containers on each build.

The docker socket is mounted internally so if docker is installed in
the container, docker can be used from the host system.
"""
from setuptools import setup, find_packages

__version__ = '0.1.1'

BASE_URL = 'https://github.com/mrasband/builder_ci'


setup(name='buildr',
      description='Simple build manifest runner, mounting the build container via docker.',
      long_description=__doc__,
      version=__version__,
      author='Matt Rasband',
      author_email='matt.rasband@gmail.com',
      license='Apache-2.0',
      url=BASE_URL,
      download_url=BASE_URL + '/archive/v' + __version__ + '.tar.gz',
      keywords=['ci', 'build'],
      packages=find_packages(),
      classifiers=[
          'Programming Language :: Python :: 3.5',
          'License :: OSI Approved :: Apache Software License',
          'Intended Audience :: Developers',
          'Development Status :: 4 - Beta',
          'Topic :: Software Development',
      ],
      setup_requires=[],
      install_requires=[
          'docker-py',
          'pyyaml',
      ],
      extras_require={},
      tests_require=[],
      entry_points={
          'console_scripts': [
              'buildr = buildr.__main__:main'
          ]
      },
      zip_safe=False)
